function [K Kd scale sums] = kernelizemaf(kernel, X, Xd, Xd2, nscale)

%
% [K Kd scale sums] = kernelizemaf(kernel, X, Xd, Xd2, nscale);
%
% Kernelize (training) data and two shifted versions and pool
% (typically for kernel MAF analysis)
%
% Input
% kernel    - kernel type, 'Gaussian' etc.
% X         - training data
% Xd        - training data, first shifted version
% Xd2       - training data, second shifted version
% nscale    - scale parameter for kernel
%
% Output
% K         - kernel for training data
% Kd        - kernel for shifted training data
% scale     - scale parameter for kernel
% sums      - aux variable for training data (symmetric)
%

% (c) Copyright 2010
% Allan Aasbjerg Nielsen
% aa@space.dtu.dk, www.imm.dtu.dk/~aa
% 15 Mar 2010

if nargin<4, error('kernelizemaf: wrong input'); end
if nargin<5, nscale = 1; end

% kernel = 'Gaussian';
% kernel = 'multiquadric';
% kernel = 'inverse multiquadric';
% kernel = 'linear, identity mapping';
% kernel = '2, second order polynomial';

[n  p ] = size(X);
[ny py] = size(Xd);
if p~=py, error('X and Xd must have same #variables'); end
[ny py] = size(Xd2);
if p~=py, error('X and Xd2 must have same #variables'); end
sums = NaN;

if ~strncmp(kernel,'l',1)
    sums = sum(X.*X,2);
    K = repmat(sums',n,1);
    K = K + K' - 2*X*X'; % squared distances beetween training obs, i.e., rows of X
    sums0 = sum(Xd.*Xd,2);
    H0 = repmat(sums0',n,1);
    Hi = repmat(sums',n,1);
    Kd  = H0 + Hi' - 2*X*Xd' ; % squared distances beween X and Xd
    clear Xd
    sums0 = sum(Xd2.*Xd2,2);
    H0 = repmat(sums0',n,1);
    Kd2 = H0 + Hi' - 2*X*Xd2'; % squared distances beween X and Xd2
    clear Xd2
	clear H0 Hi sums0
    scale = nscale*sum(real(sqrt(K(:))))/(n*(n-1)); % mean distance between obs
    scaled = scale;
end
if strncmp(kernel,'G',1)
    sigma2 = 2*scale^2;
    sigmad2 = 2*scaled^2;
    K   = exp(-K  /sigma2 ); % Gaussian
    Kd  = exp(-Kd /sigmad2);
    Kd2 = exp(-Kd2/sigmad2);
    Kd  = K - 0.5*(Kd + Kd2);
elseif strncmp(kernel,'m',1) 
    K   = sqrt(K   + scale^2); % multiquadric
    Kd  = sqrt(Kd  + scaled^2);
    Kd2 = sqrt(Kd2 + scaled^2);
    Kd  = K - 0.5*(Kd + Kd2);
elseif strncmp(kernel,'i',1) 
    K   = 1./sqrt(K   + scale^2); % inverse multiquadric
    Kd  = 1./sqrt(Kd  + scaled^2);
    Kd2 = 1./sqrt(Kd2 + scaled^2);
    Kd  = K - 0.5*(Kd + Kd2);
elseif strncmp(kernel,'l',1)
    K   = X *X'  ; % identity mapping, linear kernel
    Kd  = X *Xd' ;
    Kd2 = X *Xd2';
    Kd  = K - 0.5*(Kd + Kd2);
    scale = NaN;
else
    error('wrong kernel specification')
end
