function [mafsk vk dk idxtrain rankKd2] = kmafenvi(X, nobstrain, nmafs, lambda)

%
% Kernel maximum autocorrelation analysis, kMAF
%
% [mafsk vk dk idxtrain rankKd2] = kmafenvi(X, nobstrain, nmafs, lambda);
%
% Input
% X             - 3-D data matrix (test data)
% nobstrain     - approximate nobs for training (< ~1000)
%                 if nobstrain < 0 work on abs(nobstrain) samples only
% nmafs         - desired number of kernelized MAFs (largest autocorr)
% lambda        - amount of regularization, in interval 0 to 1
%
% Output
% mafsk         - kernel MAFs
% vk            - eigenvectors
% dk            - eigenvalues
% idxtrain      - index for training data
% rankKd2       - (numerical) rank of Kd*Kd'
%
% if nobstrain<0 we subsample without running on test data

% (c) Copyright 2008-2010
% Allan Aasbjerg Nielsen
% aa@space.dtu.dk, www.imm.dtu.dk/~aa
% 23 Sep 2010

if nargin <  3, error('kmafenvi: not enough input arguments'); end
if nargin <  4, lambda = 0; end
if nargin >  4 error('kmafenvi: too many input arguments'); end
if nargout > 5, error('kmafenvi: too many output arguments'); end

centerWithTrainingMean = 1; % as opposed: to center test data with test data mean

if centerWithTrainingMean
    h = waitbar(0,'Kernelizing training data etc...','name','Kernel MAF (1 of 4)');
else
    h = waitbar(0,'Kernelizing training data etc...','name','Kernel MAF (1 of 5)');
end

flag = 1; % kernel MAFs have variance 1 % ntrain-1
% flag = 2; % unit eigenvectors
% flag = 3; % noise part has unit variance

kernel = 'Gaussian';
% kernel = 'multiquadric';
% kernel = 'inverse multiquadric';
% kernel = 'spline, thin-plate';
% kernel = 'linear, identity mapping';
% kernel = '2, second order polynomial';

mask = NaN;
% [mask,pmask] = freadenvit('mask');
% [mask,pmask] = freadenvit('Tigalmss');
% [mask,pmask] = freadenvit('c:/aa/SSH/mask');
% [mask,pmask] = freadenvit('c:/aa/SSH/M');

seed = 4711007; % initialization of rand

readfile = 0; % input1 is 3-D matrix
if ischar(X), readfile = 1; end % input1 is filename

if readfile==1
    [X0,p] = freadenvit(X);
    nrows = p(2); ncols = p(1); nvars = p(3);
    X0 = single(X0);
else
    if ndims(X)~=3, error('X must be 3-D'); end
    [nrows,ncols,nvars] = size(X);
    X0 = single(X);
    clear X
end

if ~isnan(mask)
    if ~(nrows==pmask(1) & ncols==pmask(2))
        error('*** mask and image sizes mismatch ***');
    end
    mask = abs(mask)>0;
%     mask = imerode(mask, strel('disk',1,0));
    mask = imerode(mask, strel('square',3));
    nobsmask = sum(mask(:));
end

X0 = reshape(X0,nrows*ncols,nvars);

% (spatial) sub-sample for training, approximately nobstrain samples
rand('state',seed);
if abs(nobstrain)>nrows*ncols, nobstrain = nrows*ncols; end
idxtrain = randsample(nrows*ncols,abs(nobstrain));
if isnan(mask)
    ;
%         idxtrain = find(idxtrain);
else
    idxtrain = nonzeros(idxtrain.*mask(idxtrain));
%         idxtrain = find(idxtrain.*mask(:));
end
%
% "MAF", index runs along colmns
% remove pixels in first row and first column of X from idxtrain
idxtrain(idxtrain < nrows+1) = NaN; % first column
idxtrain(rem(idxtrain, nrows) == 1) = NaN; % first row
idxtrain = idxtrain(~isnan(idxtrain));
X  = double(X0(idxtrain,:));
Xd  = double(X0(idxtrain-1,:));
Xd2 = double(X0(idxtrain-nrows,:));
[ntrain,nvars] = size(X);
if nobstrain>0
    Xtest = X0;
    ntest = size(Xtest,1);
end
clear X0;

nscale = 1;
[K Kd scale sums] = kernelizemaf(kernel, X, Xd, Xd2, nscale);
meanK = mean(K(:));
meanrowsK = mean(K);
K = centering(K);
Kd = centering(Kd);

% MAF of kernelized training data
% max nmafs kernel MAFs output
nout = ntrain;
if nout>nmafs, nout=nmafs; end

K2 = K*K;
Kd2 = Kd*Kd';
clear Kd;
close(h);
if centerWithTrainingMean
    h = waitbar(0,'sqrtmsvd ...','name','Kernel MAF (2 of 4)');
else
    h = waitbar(0,'sqrtmsvd ...','name','Kernel MAF (2 of 5)');
end
[Bsqrt P L rankKd2] = sqrtmsvd(Kd2);
close(h);
nout = min(nout,rankKd2);
Bsqrtinv = P*diag(1./L)*P';
if centerWithTrainingMean
    h = waitbar(0,'eigs ...','name','Kernel MAF (3 of 4)');
else
    h = waitbar(0,'eigs ...','name','Kernel MAF (3 of 5)');
end
[vk,dk,flagk] = eigs(Bsqrtinv*((1-lambda)*K2+lambda*K)*Bsqrtinv, nout, 'LM');
close(h);
vk = Bsqrtinv*vk;
clear Bsqrt P L Bsqrtinv;
[dk, idx] = sort(diag(abs(dk))','descend');
vk = vk(:,idx);

if flag==1 % scale vk to give MAFs with variance 1/(ntrain-1)
    aux1 = vk'*((1-lambda)*K2+lambda*K)*vk; % dispersion of MAFs
    aux2 = 1./sqrt(diag(aux1));
    aux3 = repmat(aux2',ntrain,1);
    vk = vk.*aux3; % dispersion is now unit matrix
	vk = sqrt(ntrain-1)*vk; % variances of kMAFs are one
elseif flag==2 % vk unit vector
    aux1 = vk'*vk;
    aux2 = 1./sqrt(diag(aux1));
    aux3 = repmat(aux2',ntrain,1);
    vk = vk.*aux3; % vk now contains unit vectors in columns
    %vk'*vk
elseif flag==3 % noise part has unit variance (SMAF)
    aux1 = 0.5*vk'*Kd2*vk;
    aux2 = 1./sqrt(diag(aux1));
    aux3 = repmat(aux2',ntrain,1);
    vk = vk.*aux3; % noise dispersion is now unit matrix
    %0.5*vk'*Kd2*vk
end
clear K2 Kd2

mafsktrain = K*vk;
%cov(mafsktrain)
% sum of correlations between X and kMAFs positive
cor = corrcoef([mafsktrain X]);
cor = cor(1:nout,nout+1:end);
sumcor = sum(cor,2);
sgn = diag(sign(sumcor));
vk = vk*sgn;
mafsktrain = mafsktrain*sgn;

clear K

if nobstrain<0
    mafsk = mafsktrain;
    return
end

mafsk = NaN(ntest,nout);
if centerWithTrainingMean
    % center kernel and calculate kernelized MAFs
    h = waitbar(0, 'Kernelizing, centering and projecting test data ...', ...
        'name', 'Kernel MAF (4 of 4)');
    for rr=1:nrows
        waitbar(rr/nrows, h);
        idx = (rr-1)*ncols+1;
        idx = idx:(idx+ncols-1);
        Xk = kernelize2(kernel, X, Xtest(idx,:), scale, sums);
        Xk = Xk - repmat(meanrowsK,ncols,1)' - ...
            repmat(mean(Xk),ntrain,1) + meanK;
        mafsk(idx,:) = Xk'*vk;
    end
    close(h);
else
    % kernelize new observations which we want to feed into the model built
    % first, build up kernel means for centering
    % ... one row at a time for a large number of new observations
    h = waitbar(0, 'Kernelizing test data ...','name', 'Kernel MAF (4 of 5)');
    sumrowsK = NaN(ntest,1);
    sumcolsK = zeros(1,ntrain);
    for rr=1:nrows
        waitbar(rr/nrows, h);
        idx = (rr-1)*ncols+1;
        idx = idx:(idx+ncols-1);
        Xk = kernelize2(kernel, X, Xtest(idx,:), scale, sums);
        sumrowsK(idx) = sum(Xk',2);
        sumcolsK = sumcolsK + sum(Xk');
    end
    close(h);
    sumK = sum(sumcolsK)/(ntest*ntrain);
    sumcolsK = sumcolsK/ntest;
    sumrowsK = sumrowsK/ntrain;
    % then center kernel and calculate kernelized MAFs
    h = waitbar(0, 'Centering and projecting test data ...','name', 'Kernel MAF (5 of 5)');
    for rr=1:nrows
        waitbar(rr/nrows, h);
        idx = (rr-1)*ncols+1;
        idx = idx:(idx+ncols-1);
        Xk = kernelize2(kernel, X, Xtest(idx,:), scale, sums);
        Xk = Xk - repmat(sumcolsK,ncols,1)' - ...
            repmat(sumrowsK(idx),1,ntrain)' + sumK;
        mafsk(idx,:) = Xk'*vk;
    end
    close(h);
end

mafsk = reshape(mafsk,nrows,ncols,nout);
