function [mnfsk vk dk idxtrain rankKn2] = kmnfenvi(X, nobstrain, nmnfs, lambda)

%
% Kernel minimum noise fraction analysis, kMNF
%
% [mnfsk vk dk idxtrain rankKn2] = kmnfenvi(X, nobstrain, nmnfs, lambda)
%
% Input
% X             - 3-D data matrix (test data)
% nobstrain     - approximate nobs for training (< ~1000)
%                 if nobstrain < 0 work on abs(nobstrain) samples only
% nmnfs         - desired number of kernelized MNFs (largest SNR)
% lambda        - amount of regularization, in interval 0 to 1
%
% Output
% mnfsk         - kernel MNFs
% vk            - eigenvectors
% dk            - eigenvalues
% idxtrain      - index for training data
% rankKn2       - (numerical) rank of Kn*Kn'
%
% if nobstrain<0 we subsample without running on test data

% (c) Copyright 2008-2010
% Allan Aasbjerg Nielsen
% aa@space.dtu.dk, www.imm.dtu.dk/~aa
% 23 Sep 2010

if nargin <  3, error('kmnfenvi: not enough input arguments'); end
if nargin <  4, lambda = 0; end
if nargin >  4 error('kmnfenvi: too many input arguments'); end
if nargout > 5, error('kmnfenvi: too many output arguments'); end

centerWithTrainingMean = 1;

if centerWithTrainingMean
    h = waitbar(0,'Kernelizing training data etc...','name','Kernel MNF (1 of 4)');
else
    h = waitbar(0,'Kernelizing training data etc...','name','Kernel MNF (1 of 5)');
end

flag = 1; % kernel MNFs have variance 1 % ntrain-1
% flag = 2; % unit eigenvectors
% flag = 3; % noise part has unit variance

kernel = 'Gaussian';
% kernel = 'multiquadric';
% kernel = 'inverse multiquadric';
% kernel = 'spline, thin-plate';
% kernel = 'linear, identity mapping';
% kernel = '2, second order polynomial';
if ~strncmp(kernel,'G',1)
    error('So far only the Gaussian kernel works')
end

mask = NaN;
% [mask,pmask] = freadenvit('mask');
% [mask,pmask] = freadenvit('mask2D');
% [mask,pmask] = freadenvit('Tigalmss');
% [mask,pmask] = freadenvi('c:/aa/SSH/mask');
% % [mask,pmask] = freadenvit('c:/aa/SSH/M');

if nmnfs>30
    warning('kmnfenvi: nmnfs is larger than 30 which may require too much memory');
end

% lambda = 0;
seed = 4711007; % initialization of rand

readfile1 = 0; % input1 is 3-D matrix
if ischar(X), readfile1 = 1; end % input1 is filename

if readfile1==1
    [X0,p] = freadenvit(X);
    nrows = p(2); ncols = p(1); nvars = p(3);
    X0 = single(X0);
else
    if ndims(X)~=3, error('X must be 3-D'); end
    [nrows,ncols,nvars] = size(X);
    X0 = single(X);
    clear X
end

if ~isnan(mask)
    if ~(nrows==pmask(1) & ncols==pmask(2))
        error('kmnfenvi: mask and image sizes mismatch');
    end
    mask = abs(mask)>0;
%     mask = imerode(mask, strel('disk',1,0));
    mask = imerode(mask, strel('square',3));
    nobsmask = sum(mask(:));
end

X0 = reshape(X0,nrows*ncols,nvars);

% (spatial) sub-sample for training, approximately nobstrain samples
rand('state',seed);
if abs(nobstrain)>nrows*ncols, nobstrain = nrows*ncols; end
idxtrain = randsample(nrows*ncols,abs(nobstrain));
%idxtrain = rand(nrows*ncols,1) < abs(nobstrain)/(nrows*ncols);
if isnan(mask)
    ; %idxtrain = find(idxtrain);
else
    idxtrain = nonzeros(idxtrain.*mask(idxtrain)); ; %idxtrain = find(idxtrain.*mask(:));
end
%
% "MNF"
% remove pixels in first/last row and first/last column of X from idxtrain
idxtrain(idxtrain < nrows+1) = NaN; % first column
idxtrain(rem(idxtrain, nrows) == 1) = NaN; % first row
idxtrain(idxtrain > (ncols-1)*nrows) = NaN; % last column
idxtrain(rem(idxtrain, nrows) == 0) = NaN; % last row
idxtrain = idxtrain(~isnan(idxtrain));
X  = double(X0(idxtrain,:));
% 3 by 3 window
Xn = double(X0(idxtrain-nrows-1,:));
Xn2= double(X0(idxtrain-nrows  ,:));
Xn3= double(X0(idxtrain-nrows+1,:));
Xn4= double(X0(idxtrain      -1,:));
% 	Xn5= X;
Xn6= double(X0(idxtrain      +1,:));
Xn7= double(X0(idxtrain+nrows-1,:));
Xn8= double(X0(idxtrain+nrows  ,:));
Xn9= double(X0(idxtrain+nrows+1,:));
%
[ntrain,nvars] = size(X);
if nobstrain>0
    Xtest = X0;
    ntest = size(Xtest,1);
end
clear X0;

% while 0
nscale = 1;
[K Kn scale sums] = kernelizemnf(kernel, X, Xn, Xn2, Xn3, Xn4, Xn6, Xn7, Xn8, Xn9, nscale);
meanK = mean(K(:));
meanrowsK = mean(K);
K = centering(K);
Kn = centering(Kn);
% end %while 0

% MNF of kernelized training data
% max nmnfs kernel MNFs output
nout = ntrain;
if nout>nmnfs, nout=nmnfs; end

K2 = K*K;
Kn2 = Kn*Kn';
clear Kn;
close(h);
if centerWithTrainingMean
    h = waitbar(0,'sqrtmsvd ...','name','Kernel MNF (2 of 4)');
else
    h = waitbar(0,'sqrtmsvd ...','name','Kernel MNF (2 of 5)');
end
[Bsqrt P L rankKn2] = sqrtmsvd(Kn2);
close(h);
nout = min(nout,rankKn2);
Bsqrtinv = P*diag(1./L)*P';
if centerWithTrainingMean
    h = waitbar(0,'eigs ...','name','Kernel MNF (3 of 4)');
else
    h = waitbar(0,'eigs ...','name','Kernel MNF (3 of 5)');
end
[vk,dk,flagk] = eigs(Bsqrtinv*((1-lambda)*K2+lambda*K)*Bsqrtinv, nout, 'LM');
close(h);
vk = Bsqrtinv*vk;
clear Bsqrt P L Bsqrtinv;
[dk, idx] = sort(diag(abs(dk))','descend');
vk = vk(:,idx);

if flag==1 % scale vk to give MNFs with variance 1 % /(ntrain-1)
    aux1 = vk'*((1-lambda)*K2+lambda*K)*vk; % dispersion of MNFs
    aux2 = 1./sqrt(diag(aux1));
    aux3 = repmat(aux2',ntrain,1);
    vk = vk.*aux3; % now dispersion is unit matrix
    vk = sqrt(ntrain-1)*vk; % variances of kMNFs are one
elseif flag==2 % vk unit vector
    aux1 = vk'*vk;
    aux2 = 1./sqrt(diag(aux1));
    aux3 = repmat(aux2',ntrain,1);
    vk = vk.*aux3; % vk now contains unit vectors in columns
    %vk'*vk
elseif flag==3 % noise part has unit variance (SMNF)
    aux1 = 0.5*vk'*Kn2*vk;
    aux2 = 1./sqrt(diag(aux1));
    aux3 = repmat(aux2',ntrain,1);
    vk = vk.*aux3; % now noise dispersion is unit matrix
    %0.5*vk'*Kn2*vk
end
clear K2 Kn2

mnfsktrain = K*vk;
%cov(mnfsktrain)
% sum of correlations between X and kMNFs positive
cor = corrcoef([mnfsktrain X]);
cor = cor(1:nout,nout+1:end);
sumcor = sum(cor,2);
sgn = diag(sign(sumcor));
vk = vk*sgn;
mnfsktrain = mnfsktrain*sgn;

clear K

if nobstrain<0
    mnfsk = mnfsktrain;
    return
end

mnfsk = NaN(ntest,nout);
if centerWithTrainingMean
    % center kernel and calculate kernelized MNFs
    h = waitbar(0, 'Kernelizing, centering and projecting test data ...', ...
        'name', 'Kernel MNF (4 of 4)');
    for rr=1:nrows
        waitbar(rr/nrows, h);
        idx = (rr-1)*ncols+1;
        idx = idx:(idx+ncols-1);
        Xk = kernelize2(kernel, X, Xtest(idx,:), scale, sums);
        Xk = Xk - repmat(meanrowsK,ncols,1)' - ...
            repmat(mean(Xk),ntrain,1) + meanK;
        mnfsk(idx,:) = Xk'*vk;
    end
    close(h);
else
    % kernelize new observations which we want to feed into the model built
    % first, build up kernel means for centering
    % ... one row at a time for a large number of new observations
    h = waitbar(0, 'Kernelizing test data ...','name', 'Kernel MNF (4 of 5)');
    sumrowsK = NaN(ntest,1);
    sumcolsK = zeros(1,ntrain);
    for rr=1:nrows
        waitbar(rr/nrows, h);
        idx = (rr-1)*ncols+1;
        idx = idx:(idx+ncols-1);
        Xk = kernelize2(kernel, X, Xtest(idx,:), scale, sums);
        sumrowsK(idx) = sum(Xk',2);
        sumcolsK = sumcolsK + sum(Xk');
    end
    close(h);
    sumK = sum(sumcolsK)/(ntest*ntrain);
    sumcolsK = sumcolsK/ntest;
    sumrowsK = sumrowsK/ntrain;
    % then center kernel and calculate kernelized MNFs
    h = waitbar(0, 'Centering and projecting test data ...','name', 'Kernel MNF (5 of 5)');
    for rr=1:nrows
        waitbar(rr/nrows, h);
        idx = (rr-1)*ncols+1;
        idx = idx:(idx+ncols-1);
        Xk = kernelize2(kernel, X, Xtest(idx,:), scale, sums);
        Xk = Xk - repmat(sumcolsK,ncols,1)' - ...
            repmat(sumrowsK(idx),1,ntrain)' + sumK;
        mnfsk(idx,:) = Xk'*vk;
    end
    close(h);
end

mnfsk = reshape(mnfsk,nrows,ncols,nout);
