function [pcsk vk dk idxtrain] = kpcaenvi(X, nobstrain, npcs, lambda)

%
% Kernel principal component analysis, kPCA
%
% [pcsk vk dk idxtrain] = kpcaenvi(X, nobstrain, npcs, lambda);
%
% Input
% X             - 3-D input matrix, ncols by nrows by nvars
% nobstrain     - approximate nobs for training (< ~1000)
%                 if nobstrain < 0 work on abs(nobstrain) samples only
% npcs          - desired number of kernelized PCs (largest eigen values)
% lambda        - amount of regularization, in interval 0 to 1
%
% Output
% pcsk          - kernelized PCs
% vk            - eigenvectors divided by square root of corresponding
%                 eigenvalues
% dk            - eigenvalues
% idxtrain      - index for training data
%
% if nobstrain<0 we subsample without running on test data

% (c) Copyright 2008-2010
% Allan Aasbjerg Nielsen
% aa@space.dtu.dk, www.imm.dtu.dk/~aa
% 23 Sep 2010

if nargin <  3, error('kpcaenvi: not enough input arguments'); end
if nargin <  4, lambda = 0; end
if nargin >  4, error('kpcaenvi: too many input arguments'); end
if nargout > 4, error('kpcaenvi: too many output arguments'); end

centerWithTrainingMean = 1;

if centerWithTrainingMean
    h = waitbar(0,'Kernelizing training data etc...','name','Kernel PCA (1 of 3)');
else
    h = waitbar(0,'Kernelizing training data etc...','name','Kernel PCA (1 of 4)');
end

kernel = 'Gaussian';
% kernel = 'multiquadric';
% kernel = 'inverse multiquadric';
% kernel = 'spline, thin-plate';
% kernel = 'linear, identity mapping';
% kernel = '2, second order polynomial'; 

mask = NaN;
% [mask,pmask] = freadenvi('c:/aa/SSH/mask');
% [mask,pmask] = freadenvit('c:/aa/SSH/M');

seed = 4711007; % initialization of rand

readfile = 0; % input1 is 3-D matrix
if ischar(X), readfile = 1; end % input1 is filename

if readfile==1
    [X0,p] = freadenvit(X);
    nrows = p(2); ncols = p(1); nvars = p(3);
    X0 = single(X0);
else
    if ndims(X)~=3, error('X must be 3-D'); end
    [nrows,ncols,nvars] = size(X);
    X0 = single(X);
    clear X
end

if ~isnan(mask)
    if ~(nrows==pmask(1) & ncols==pmask(2))
%     if ~(nrows==pmask(2) & ncols==pmask(1))
        error('*** mask and image sizes mismatch ***');
    end
    mask = abs(mask)>0;
    nobsmask = sum(mask(:));
    nobstrain = round(nobstrain*nrows*ncols/nobsmask);
end

X0 = reshape(X0,nrows*ncols,nvars);

% (spatial) sub-sample for training, approximately nobstrain samples
rand('state',seed);
if abs(nobstrain)>nrows*ncols, nobstrain = nrows*ncols, end
idxtrain = randsample(nrows*ncols,abs(nobstrain));
%idxtrain = rand(nrows*ncols,1) < abs(nobstrain)/(nrows*ncols);
if isnan(mask)
    ; %idxtrain = find(idxtrain);
else
    idxtrain = nonzeros(idxtrain.*mask(idxtrain));
    %idxtrain = find(idxtrain.*mask(:));
end
X = double(X0(idxtrain,:));
[ntrain,nvars] = size(X);
if nobstrain>0
    Xtest = X0;
    ntest = size(Xtest,1);
end
clear X0;

nscale = 1;
[K scale sums] = kernelize(kernel, X, nscale);
meanK = mean(K(:));
meanrowsK = mean(K);
K = centering(K);

% PCA of kernelized training data
% max npcs kernelized PCs output
nout = ntrain;
if nout>npcs, nout=npcs; end
close(h);
if centerWithTrainingMean
    h = waitbar(0,'eigs ...','name','Kernel PCA (2 of 3)');
else
    h = waitbar(0,'eigs ...','name','Kernel PCA (2 of 4)');
end
[vk,dk,flagk] = eigs((1-lambda)*K+lambda*eye(ntrain), nout, 'LM');
close(h);
if flagk~=0, warning('*** Convergence problems in eigs ***'); end
dk = diag(abs(dk))';
vk = sqrt(ntrain-1)*vk*diag(1./sqrt(dk));

pcsktrain = K*vk;
clear K
% sum of correlations between X and kPCs positive
cor = corrcoef([pcsktrain X]);
cor = cor(1:nout,nout+1:end);
sumcor = sum(cor,2);
sgn = diag(sign(sumcor));
vk = vk*sgn;
pcsktrain = pcsktrain*sgn;

if nobstrain<0
    pcsk = pcsktrain;
    return
end

pcsk = NaN(ntest,nout);
if centerWithTrainingMean
    % center kernel and calculate kernelized PCs
    h = waitbar(0, 'Kernelizing, centering and projecting test data ...', ...
        'name', 'Kernel PCA (3 of 3)');
    for rr=1:nrows
        waitbar(rr/nrows, h);
        idx = (rr-1)*ncols+1;
        idx = idx:(idx+ncols-1);
        Xk = kernelize2(kernel, X, Xtest(idx,:), scale, sums);
        Xk = Xk - repmat(meanrowsK,ncols,1)' - ...
            repmat(mean(Xk),ntrain,1) + meanK;
        pcsk(idx,:) = Xk'*vk;
    end
    close(h);
else
    % kernelize new observations which we want to feed into the model built
    % first, build up kernel means for centering
    % ... one row at a time for a large number of new observations
    h = waitbar(0, 'Kernelizing test data ...','name', 'Kernel PCA (3 of 4)');
    sumrowsK = NaN(ntest,1);
    sumcolsK = zeros(1,ntrain);
    for rr=1:nrows
        waitbar(rr/nrows, h);
        idx = (rr-1)*ncols+1;
        idx = idx:(idx+ncols-1);
        Xk = kernelize2(kernel, X, Xtest(idx,:), scale, sums);
        sumrowsK(idx) = sum(Xk',2);
        sumcolsK = sumcolsK + sum(Xk');
    end
    close(h);
    sumK = sum(sumcolsK)/(ntest*ntrain);
    sumcolsK = sumcolsK/ntest;
    sumrowsK = sumrowsK/ntrain;
    % then center kernel and calculate kernelized PCs
    h = waitbar(0, 'Centering and projecting test data ...','name', 'Kernel PCA (4 of 4)');
    for rr=1:nrows
        waitbar(rr/nrows, h);
        idx = (rr-1)*ncols+1;
        idx = idx:(idx+ncols-1);
        Xk = kernelize2(kernel, X, Xtest(idx,:), scale, sums);
        Xk = Xk - repmat(sumcolsK,ncols,1)' - ...
            repmat(sumrowsK(idx),1,ntrain)' + sumK;
        pcsk(idx,:) = Xk'*vk;
    end
    close(h);
end

pcsk = reshape(pcsk,nrows,ncols,nout);
